#include <fstream.h>
#include <stdlib.h>
#include <alloc.h>
#include <conio.h>
#include <time.h>
#include <dos.h>
#include "tables.h"

#define FILENAME      "texture.bmp"
#define R             0x0000040L
#define PERSPECT      0x0000100L
#define PERSPECT1     0x0000100L
#define TILE          9
#define TEX_SIZE      16384
#define LEVELS        32
#define LEVELBITS     5
#define MIN_SCALE     0x00080000L
#define MAX_SCALE     0x00010000L
#define SCALE_STEP    0x00004000L
#define MAX_DELAY     100
#define MAX_COLORS    192

unsigned char *tex;
unsigned char *map;
unsigned char pal[1024];
void far *buf;
long nz;

void SetGraphMode();
void SetTextMode();
void SetColor(unsigned char, unsigned char, unsigned char, unsigned char);
void LoadBMP(char *);
void SetPalette();
void Background();
void CopyBuf();
int FindColor(unsigned char, unsigned char, unsigned char);
void MatchPalette();

struct Vector {
  long x, y, z;
  Vector() { }
  Vector(long x_, long y_, long z_)
  {
    x = x_;
    y = y_;
    z = z_;
  }
  Vector(long v)
  {
    x = y = z = v;
  }
  Vector friend operator ^(Vector, Vector);
  Vector friend operator +(Vector, Vector);
  Vector friend operator -(Vector, Vector);
  Vector& operator /=(Vector d)
  {
    x = (x << 16) / d.x;
    y = (y << 16) / d.y;
    z = (z << 16) / d.z;
    return *this;
  }
  Vector& operator +=(Vector v)
  {
    x += v.x;
    y += v.y;
    z += v.z;
    return *this;
  }
  Vector& operator -=(Vector v)
  {
    x -= v.x;
    y -= v.y;
    z -= v.z;
    return *this;
  }
  bool friend operator >(Vector, double);
  bool friend operator <(Vector, double);
};

Vector operator ^(Vector v1, Vector v2)
{
  return Vector(v1.y * v2.z - v1.z * v2.y,
                v1.z * v2.x - v1.x * v2.z,
                v1.x * v2.y - v1.y * v2.x);
}

Vector operator +(Vector v1, Vector v2)
{
  return Vector(v1.x + v2.x, v1.y + v2.y, v1.z + v2.z);
}

Vector operator -(Vector v1, Vector v2)
{
  return Vector(v1.x + v2.x, v1.y + v2.y, v1.z + v2.z);
}

bool operator >(Vector v, double d)
{
  if (v.x <= v.y && v.x <= v.z)
    return (v.x > d);
  else if (v.y <= v.x && v.y <= v.z)
    return (v.y > d);
  return (v.z > d);
}

bool operator <(Vector v, double d)
{
  if (v.x >= v.y && v.x >= v.z)
    return (v.x < d);
  else if (v.y >= v.x && v.y >= v.z)
    return (v.y < d);
  return (v.z < d);
}

void Draw1stTriangle(long, long, long, long, long, Vector, Vector, Vector);
void DrawTriangle(Vector, Vector, Vector, bool);
Vector e1, e2, a, l, m, n;

class Edge {
    Vector v1, v2, v3, v4;
    Vector o1, o2, o3, o4;
    Vector n, on;
    Vector RotateZ(Vector& s, int a)
    {
      return Vector((s.x * cos_table[a] + s.y * sin_table[a]) >> 16,
                    (s.y * cos_table[a] - s.x * sin_table[a]) >> 16, s.z);
    }
    Vector RotateX(Vector& s, int a)
    {
      return Vector(s.x, (s.y * cos_table[a] - s.z * sin_table[a]) >> 16,
                    (s.z * cos_table[a] + s.y * sin_table[a]) >> 16);
    }
    Vector RotateY(Vector& s, int a)
    {
      return Vector((s.x * cos_table[a] + s.z * sin_table[a]) >> 16,
                    s.y, (s.z * cos_table[a] - s.x * sin_table[a]) >> 16);
    }
  public:
    Edge() { }
    Edge(Vector p1, Vector p2, Vector p3, Vector p4, Vector normal)
    {
      v1 = o1 = p1;
      v2 = o2 = p2;
      v3 = o3 = p3;
      v4 = o4 = p4;
      on = n = normal;
    }
    void RotateX(int a)
    {
      v1 = RotateX(v1, a);
      v2 = RotateX(v2, a);
      v3 = RotateX(v3, a);
      v4 = RotateX(v4, a);
      n = RotateX(n, a);
    }
    void RotateY(int a)
    {
      v1 = RotateY(v1, a);
      v2 = RotateY(v2, a);
      v3 = RotateY(v3, a);
      v4 = RotateY(v4, a);
      n = RotateY(n, a);
    }
    void RotateZ(int a)
    {
      v1 = RotateZ(v1, a);
      v2 = RotateZ(v2, a);
      v3 = RotateZ(v3, a);
      v4 = RotateZ(v4, a);
      n = RotateZ(n, a);
    }
    void Scale(Vector d)
    {
      v1 /= d;
      v2 /= d;
      v3 /= d;
      v4 /= d;
    }
    void Draw()
    {
      if (n.z > 0)
      {
        Vector o(0, 0, (R << 2) + R);
        nz = (n.z - 1) << 8;
        DrawTriangle(v2 + o, v3 + o, v1 + o, true);
        DrawTriangle(v4 + o, v1 + o, v3 + o, false);
      }
      v1 = o1;
      v2 = o2;
      v3 = o3;
      v4 = o4;
      n = on;
    }
    long AverageZ()
    {
      return v1.z + v2.z + v3.z + v4.z;
    }
};

void Draw1stTriangle(long x1, long x2, long x3, long y1, long y2, Vector n, Vector m, Vector l)
{
  long deltay = y2 - y1, u1, v1, u2, v2, u, v, d;
  long xx1 = x1 << 16, xx2 = x1 << 16, X1, X2, Y;
  long du, dv, t1, t2, t3, xx3, xx4, dx;
  int i = y1, j, dy = 1;
  unsigned char c = 0x01;

  if (!deltay) return;

  if (deltay < 0)
  {
    dy = -1;
    deltay = -deltay;
  }

  long dx1 = ((x2 - x1) << 16) / deltay;
  long dx2 = ((x3 - x1) << 16) / deltay;

  for (; dy > 0 ? (i <= y2) : (i >= y2); i += dy)
  {
    xx3 = (xx1 >> 16);
    xx4 = (xx2 >> 16);
    dx = xx4 - xx3;
    X1 = xx3 * PERSPECT1;
    X2 = xx4 * PERSPECT1;
    Y = long(i) * PERSPECT1;

    t1 = Y * n.y + n.z;
    t2 = Y * m.y + m.z;
    t3 = Y * l.y + l.z;
    d = (X1 * n.x + t1) >> 16;

    if (d)
    {
      u1 = ((X1 * m.x + t2) >> TILE) / d;
      v1 = ((X1 * l.x + t3) >> TILE) / d;
      d = (X2 * n.x + t1) >> 16;
      if (d)
      {
        u2 = ((X2 * m.x + t2) >> TILE) / d;
        v2 = ((X2 * l.x + t3) >> TILE) / d;
      }
    }

    u = u1 << 16;
    v = v1 << 16;

    if (dx)
    {
      du = ((u2 - u1) << 16) / dx;
      dv = ((v2 - v1) << 16) / dx;
    }

    for (j = xx3; j <= xx4; j++)
    {
      c = map[nz | tex[(u >> 9) & 0x3F80 | (v >> 16) & 0x7F]];
      asm  les    di,  buf
      asm  mov    bx,  i
      asm  add    bx,  100
      asm  mov    ax,  bx
      asm  shl    bx,  8
      asm  shl    ax,  6
      asm  add    bx,  ax
      asm  add    bx,  j
      asm  add    bx,  160
      asm  add    di,  bx
      asm  mov    al,  c
      asm  stosb
      u += du;
      v += dv;
    }
    xx1 += dx1;
    xx2 += dx2;
  }
}

void DrawTriangle(Vector v1, Vector v2, Vector v3, bool thesame)
{
  long x1 = v1.x, x2 = v2.x, x3 = v3.x;
  long y1 = v1.y, y2 = v2.y, y3 = v3.y;
  if (thesame)
  {
    e1 = Vector(x2 - x1, y2 - y1, v2.z - v1.z);
    e2 = Vector(x3 - x1, y3 - y1, v3.z - v1.z);
    a = Vector(v1.x, v1.y, v1.z);
    l = Vector(a ^ e1);
    m = Vector(e2 ^ a);
    n = Vector(e1 ^ e2);

    l.z <<= 16;
    m.z <<= 16;
    n.z <<= 16;
  }
  long xx1, xx3, xx4, dx2;
  long dy;

  x1 = PERSPECT * x1 / v1.z;
  y1 = PERSPECT * y1 / v1.z;
  x2 = PERSPECT * x2 / v2.z;
  y2 = PERSPECT * y2 / v2.z;
  x3 = PERSPECT * x3 / v3.z;
  y3 = PERSPECT * y3 / v3.z;

  if (y1 == y2 && y1 == y3) return;
  if (x1 == x2 && x1 == x3) return;

  if (y2 < y1)
  {
    y2 ^= y1; y1 ^= y2; y2 ^= y1;
    x2 ^= x1; x1 ^= x2; x2 ^= x1;
  }

  if (y3 < y1)
  {
    y3 ^= y1; y1 ^= y3; y3 ^= y1;
    x3 ^= x1; x1 ^= x3; x3 ^= x1;
  }

  if (y3 < y2)
  {
    y3 ^= y2; y2 ^= y3; y3 ^= y2;
    x3 ^= x2; x2 ^= x3; x3 ^= x2;
  }

  if (y2 == y3)
  {
    if (x2 > x3)
      Draw1stTriangle(x1, x3, x2, y1, y2, n, m, l);
    else
      Draw1stTriangle(x1, x2, x3, y1, y2, n, m, l);
  }
  else
  {
    dy = y3 - y1;
    if (dy < 0) dy = -dy;
    xx1 = x1;
    xx3 = x3;
    xx1 <<= 16;
    xx3 <<= 16;
    dx2 = (xx3 - xx1) / dy;
    dy = y2 - y1;
    if (dy < 0) dy = -dy;
    xx4 = (dx2 * dy) >> 16;
    xx4 += x1;
    if (xx4 < x2)
    {
      Draw1stTriangle(x1, xx4, x2, y1, y2, n, m, l);
      Draw1stTriangle(x3, xx4, x2, y3, y2, n, m, l);
    }
    else
    {
      Draw1stTriangle(x1, x2, xx4, y1, y2, n, m, l);
      Draw1stTriangle(x3, x2, xx4, y3, y2, n, m, l);
    }
  }
}

void SetGraphMode()
{
  asm  mov    ax,  13h
  asm  int    10h
}

void SetTextMode()
{
  asm  mov    ax,  03h
  asm  int    10h
}

void SetColor(unsigned char color, unsigned char r, unsigned char g, unsigned char b)
{
  asm  mov    dx,  3C8h
  asm  mov    al,  color
  asm  out    dx,  al
  asm  inc    dx
  asm  mov    al,  r
  asm  shr    al,  2
  asm  out    dx,  al
  asm  mov    al,  g
  asm  shr    al,  2
  asm  out    dx,  al
  asm  mov    al,  b
  asm  shr    al,  2
  asm  out    dx,  al
}

void SetPalette()
{
  for (int i = 1; i < 192; i++)
    SetColor(i, pal[(i << 2) + 2], pal[(i << 2) + 1], pal[i << 2]);
  for (int i = 192; i < 256; i++)
    SetColor(i, 0, 0, (i - 192) << 2);
}

void LoadBMP(char *filename)
{
  ifstream in(filename, ios::binary);
  if (!in)
  {
    cout << "File \"" << FILENAME << "\" not found\n";
    exit(1);
  }
  in.seekg(0x36, ios::beg);
  in.read(pal, sizeof(pal));
  in.read(tex, TEX_SIZE);
}

void Background()
{
  asm  les    di,  buf
  asm  xor    ax,  ax
  asm  mov    cx,  50
l1:
  asm  push   cx
  asm  mov    al,  cl
  asm  add    al,  192
  asm  mov    ah,  al
  asm  mov    cx,  640
  asm  rep    stosw
  asm  pop    cx
  asm  loop   l1
}

void CopyBuf()
{
  asm  push   ds
  asm  push   0A000h
  asm  pop    es
  asm  xor    di,  di
  asm  lds    si,  buf
  asm  mov    cx,  32000
  asm  rep    movsw
  asm  pop    ds
}

int FindColor(unsigned char r2, unsigned char g2, unsigned char b2)
{
  long r1, g1, b1, min = 0x7FFFFFFFL, d;
  int n, i = 0;

  for (; i < MAX_COLORS; i++)
  {
    r1 = pal[(i << 2) + 2];
    g1 = pal[(i << 2) + 1];
    b1 = pal[i << 2];
    d = (r1 - r2) * (r1 - r2) + (g1 - g2) * (g1 - g2) + (b1 - b2) * (b1 - b2);
    if (d < min)
    {
      min = d;
      n = i;
    }
  }

  return n;
}

void MatchPalette()
{
  for (int i = 0; i < LEVELS; i++)
    for (int j = 0; j < MAX_COLORS; j++)
      map[(i << 8) + j] = FindColor((pal[(j << 2) + 2] * (i + 1)) >> LEVELBITS,
                                    (pal[(j << 2) + 1] * (i + 1)) >> LEVELBITS,
                                    (pal[(j << 2)] * (i + 1)) >> LEVELBITS);
}

void main()
{
  long end = 0;
  int a = 0, i, j, n, speed = 0;
  char c;
  long frames = 0;
  Edge e[6];
  long z[6], max;
  Vector scale(0x00010000L);
  Vector n1(0, 0, -LEVELS);
  Vector n2(0, 0, LEVELS);
  Vector n3(0, LEVELS, 0);
  Vector n4(0, -LEVELS, 0);
  Vector n5(-LEVELS, 0, 0);
  Vector n6(LEVELS, 0, 0);
  e[0] = Edge(Vector(-R, R, R), Vector(R, R, R), Vector(R, -R, R), Vector(-R, -R, R), n1);
  e[1] = Edge(Vector(-R, R, -R), Vector(R, R, -R), Vector(R, -R, -R), Vector(-R, -R, -R), n2);
  e[2] = Edge(Vector(-R, -R, R), Vector(-R, -R, -R), Vector(R, -R, -R), Vector(R, -R, R), n3);
  e[3] = Edge(Vector(-R, R, R), Vector(-R, R, -R), Vector(R, R, -R), Vector(R, R, R), n4);
  e[4] = Edge(Vector(R, R, -R), Vector(R, R, R), Vector(R, -R, R), Vector(R, -R, -R), n5);
  e[5] = Edge(Vector(-R, R, -R), Vector(-R, R, R), Vector(-R, -R, R), Vector(-R, -R, -R), n6);

  buf = farmalloc(64000L);
  tex = new unsigned char[TEX_SIZE];
  map = new unsigned char[LEVELS << 8];
  LoadBMP(FILENAME);

  SetGraphMode();
  SetPalette();
  MatchPalette();

  do
  {
    Background();
    for (i = 0; i < 6; i++)
    {
      e[i].Scale(scale);
      e[i].RotateY(a);
      e[i].RotateX(a);
      z[i] = e[i].AverageZ();
    }
    for (i = 0; i < 6; i++)
    {
      max = -0x10000L;
      for (j = 0; j < 6; j++)
        if (z[j] > max)
        {
          max = z[j];
          n = j;
        }
      e[n].Draw();
      z[n] = -0x10000L;
    }
    CopyBuf();
    a++;
    a %= 360;
    frames++;
    if (speed)
      delay(speed);
    if (kbhit())
    {
      while (kbhit())
      {
        c = getch();
        switch (c)
        {
          case 83: if (scale.x < MIN_SCALE) scale.x += SCALE_STEP; break;
          case 82: if (scale.x > MAX_SCALE) scale.x -= SCALE_STEP; break;
          case 79: if (scale.y < MIN_SCALE) scale.y += SCALE_STEP; break;
          case 71: if (scale.y > MAX_SCALE) scale.y -= SCALE_STEP; break;
          case 81: if (scale.z < MIN_SCALE) scale.z += SCALE_STEP; break;
          case 73: if (scale.z > MAX_SCALE) scale.z -= SCALE_STEP; break;
          case 80: if (scale < MIN_SCALE) scale += SCALE_STEP; break;
          case 72: if (scale > MAX_SCALE) scale -= SCALE_STEP; break;
          case '-': if (speed < MAX_DELAY) speed++; break;
          case '+': if (speed) speed--;
        }
      }
    }
  } while (c != 27);

  delete map;
  delete tex;
  farfree(buf);

  end = clock() / CLK_TCK;
  SetTextMode();
  cout << frames << " frames, " << end << " seconds - " << (frames / end) << " FPS\n";
  cprintf("Texture mapping - by HackMaster\n");
}
